import sys

import PySide6.QtGui as QtGui
import PySide6.QtWidgets as QtWidgets
import PySide6.QtCore as QtCore

from . import value_table
from . import key_tree
from . import hive_info_table
from . import license_dialog
from . import find_dialog
from . import helpers


# Set the app ID on windows (helps with making sure icon is used)
try:
    import ctypes
    appid = "ajsmith.registryspy.desktop.v1"
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(appid)
except (AttributeError, OSError):
    pass


class RegViewer(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()

        self.app = QtWidgets.QApplication.instance()
        self.settings = QtCore.QSettings()

        self.initial_style = self.app.style().name()

        use_native_style = self.settings.value(
            "view/native_style", False, bool)

        if not use_native_style:
            self.app.setStyle("fusion")

        # Set up main window
        self.resize(1200, 700)
        self.setWindowTitle(helpers.APP_NAME)

        # Restore window geometry from settings
        self.restoreGeometry(self.settings.value(
            "view/geometry", QtCore.QByteArray()))

        self.tree = key_tree.KeyTree(self)
        self.find_dialog = find_dialog.FindDialog(self)

        # Set up file menu
        file_menu = QtWidgets.QMenu("&File", self)
        open_action = QtGui.QAction("Open Hive", self)
        open_action.setShortcut(QtGui.QKeySequence.Open)
        open_action.triggered.connect(self.show_open_file)
        file_menu.addAction(open_action)
        close_action = QtGui.QAction("Close Selected Hive", self)
        close_action.setShortcut(QtGui.QKeySequence(
            QtCore.Qt.SHIFT | QtCore.Qt.Key_Delete))
        close_action.triggered.connect(self.tree.remove_selected_hive)
        file_menu.addAction(close_action)
        close_all_action = QtGui.QAction("Close All Hives", self)
        close_all_action.setShortcut(QtGui.QKeySequence(
            QtCore.Qt.CTRL | QtCore.Qt.SHIFT | QtCore.Qt.Key_Delete))
        close_all_action.triggered.connect(self.tree.remove_all_hives)
        file_menu.addAction(close_all_action)
        file_menu.addSeparator()
        quit_action = QtGui.QAction("Quit", self)
        quit_action.triggered.connect(self.close)
        file_menu.addAction(quit_action)
        self.menuBar().addMenu(file_menu)

        # Set up find menu
        find_menu = QtWidgets.QMenu("Fin&d", self)
        find_action = QtGui.QAction("Find...", self)
        find_action.setShortcut(QtGui.QKeySequence.Find)
        find_action.triggered.connect(self.show_find)
        find_menu.addAction(find_action)
        find_next_action = QtGui.QAction("Find Next", self)
        find_next_action.setShortcut(QtGui.QKeySequence.FindNext)
        find_next_action.triggered.connect(self.find_dialog.handle_find)
        find_menu.addAction(find_next_action)
        # find_previous_action = QtGui.QAction("Find Previous", self)
        # find_previous_action.setShortcut(QtGui.QKeySequence.FindPrevious)
        # find_menu.addAction(find_previous_action)
        self.menuBar().addMenu(find_menu)

        # Set up view menu
        view_menu = QtWidgets.QMenu("&View", self)
        self.native_style_action = QtGui.QAction("Use native style", self)
        self.native_style_action.setCheckable(True)
        self.native_style_action.setChecked(use_native_style)
        self.native_style_action.toggled.connect(self.toggle_style)
        view_menu.addAction(self.native_style_action)
        self.menuBar().addMenu(view_menu)

        # Set up help menu
        help_menu = QtWidgets.QMenu("&Help", self)
        about_action = QtGui.QAction("About", self)
        about_action.triggered.connect(self.show_about)
        help_menu.addAction(about_action)
        licenses_action = QtGui.QAction("Third-Party Licenses", self)
        licenses_action.triggered.connect(self.show_licenses)
        help_menu.addAction(licenses_action)
        self.menuBar().addMenu(help_menu)

        self.menuBar().setContextMenuPolicy(QtCore.Qt.PreventContextMenu)

        self.statusBar()

        # Set up toolbar
        toolbar = QtWidgets.QToolBar("Main", self)
        toolbar.setMovable(False)
        toolbar.setFloatable(False)
        toolbar.toggleViewAction().setEnabled(False)
        toolbar.setContextMenuPolicy(QtCore.Qt.PreventContextMenu)
        open_action = QtGui.QAction(
            QtGui.QIcon(helpers.resource_path("img/open.png")), "Open Hive", toolbar)
        open_action.triggered.connect(self.show_open_file)
        toolbar.addAction(open_action)
        close_action = QtGui.QAction(
            QtGui.QIcon(helpers.resource_path("img/close.png")), "Close Selected Hive", toolbar)
        close_action.triggered.connect(self.tree.remove_selected_hive)
        toolbar.addAction(close_action)
        close_all_action = QtGui.QAction(
            QtGui.QIcon(helpers.resource_path("img/close_all.png")), "Close All Hives", toolbar)
        close_all_action.triggered.connect(self.tree.remove_all_hives)
        toolbar.addAction(close_all_action)
        find_action = QtGui.QAction(
            QtGui.QIcon(helpers.resource_path("img/find.png")), "Find", toolbar)
        find_action.triggered.connect(self.show_find)
        toolbar.addAction(find_action)
        find_next_action = QtGui.QAction(
            QtGui.QIcon(helpers.resource_path("img/find_next.png")), "Find Next", toolbar)
        find_next_action.triggered.connect(self.find_dialog.handle_find)
        toolbar.addAction(find_next_action)
        self.addToolBar(toolbar)

        # Set up main layout
        main_widget = QtWidgets.QWidget(self)
        main_layout = QtWidgets.QVBoxLayout(main_widget)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_widget.setLayout(main_layout)

        self.value_table = value_table.ValueTable(main_widget)

        self.uri_textbox = QtWidgets.QLineEdit(main_widget)
        self.uri_textbox.returnPressed.connect(self.tree.handle_uri_change)
        main_layout.addWidget(self.uri_textbox)

        self.statusBar().setStyleSheet(
            "QStatusBar QLabel { border-color: lightgray; border-style: solid; border-width: 0 1px 0 0; }")

        tree_container = QtWidgets.QWidget()
        tree_container_layout = QtWidgets.QVBoxLayout(tree_container)
        tree_container_layout.setContentsMargins(0, 0, 0, 0)
        tree_container.setLayout(tree_container_layout)

        tree_container_layout.addWidget(self.tree)
        self.tree.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        self.hive_info = hive_info_table.HiveInfoTable(tree_container)
        hive_geometry = self.hive_info.geometry()
        hive_geometry.setHeight(5)
        self.hive_info.setGeometry(hive_geometry)
        self.hive_info.setMinimumHeight(10)
        self.hive_info.setSizeAdjustPolicy(
            self.hive_info.SizeAdjustPolicy.AdjustToContents)
        tree_container_layout.addWidget(self.hive_info)

        value_splitter = QtWidgets.QSplitter(
            QtGui.Qt.Orientation.Vertical)
        value_splitter.addWidget(self.value_table)
        self.value_hex = QtWidgets.QPlainTextEdit()
        mono_font = QtGui.QFont()
        mono_font.setFamilies(["Courier New", "Monospaced"])
        mono_font.setStyleHint(QtGui.QFont.Monospace)
        self.value_hex.setFont(mono_font)
        self.value_hex.setLineWrapMode(self.value_hex.LineWrapMode.NoWrap)
        value_splitter.addWidget(self.value_hex)
        value_splitter.setStretchFactor(0, 2)
        value_splitter.setStretchFactor(1, 1)

        main_splitter = QtWidgets.QSplitter(main_widget)
        main_splitter.addWidget(tree_container)
        main_splitter.addWidget(value_splitter)
        main_splitter.setStretchFactor(0, 4)
        main_splitter.setStretchFactor(1, 5)
        main_splitter.setChildrenCollapsible(False)
        main_layout.addWidget(main_splitter)
        self.setCentralWidget(main_widget)

        self.progress_bar = QtWidgets.QProgressBar(self.statusBar())
        self.progress_bar.setMaximumWidth(100)
        self.progress_bar.hide()
        self.statusBar().addPermanentWidget(self.progress_bar)

    def show_about(self):
        QtWidgets.QMessageBox().about(
            self, f"About {helpers.APP_NAME}", helpers.ABOUT_TEXT)

    def show_licenses(self):
        license = license_dialog.LicenseDialog(self)
        license.exec()

    def show_find(self):
        self.find_dialog.exec()

    def show_open_file(self):
        """Show the open file dialog"""
        file_selection = QtWidgets.QFileDialog.getOpenFileNames(
            self, "Open Registry File")

        self.statusBar().showMessage("Loading...")
        self.statusBar().repaint()

        if isinstance(file_selection, tuple) and len(file_selection) > 0 and len(file_selection[0]) > 0:
            for file in file_selection[0]:
                self.open_file(file)

        self.statusBar().clearMessage()

    def open_file(self, filename: str):
        self.tree.load_hive(filename)

    def toggle_style(self):
        if self.native_style_action.isChecked():
            self.app.setStyle(self.initial_style)
            self.settings.setValue("view/native_style", True)
        else:
            self.app.setStyle("fusion")
            self.settings.setValue("view/native_style", False)

    def closeEvent(self, event):
        """Save the current geometry of the application"""
        self.settings.setValue("view/geometry", self.saveGeometry())
        event.accept()


def main():
    app = QtWidgets.QApplication(sys.argv)

    app.setOrganizationName(helpers.ORGANIZATION_NAME)
    app.setOrganizationDomain(helpers.ORGANIZATION_DOMAIN)
    app.setApplicationName(helpers.APP_NAME)

    # f = QtCore.QFile("qdarkstyle/dark/style.qss")
    # f.open(QtCore.QFile.ReadOnly | QtCore.QFile.Text)
    # qss = QtCore.QTextStream(f)
    # app.setStyleSheet(qss.readAll())

    # Dark Mode
    # https://github.com/mguludag/QEasySettings/blob/main/qeasysettings.cpp
    # palette = QtGui.QPalette()
    # palette.setColor(QtGui.QPalette.Window, QtGui.QColor(53, 53, 53))
    # palette.setColor(QtGui.QPalette.WindowText, QtCore.Qt.white)
    # palette.setColor(QtGui.QPalette.Base, QtGui.QColor(25, 25, 25))
    # palette.setColor(QtGui.QPalette.AlternateBase, QtGui.QColor(53, 53, 53))
    # palette.setColor(QtGui.QPalette.ToolTipBase, QtGui.QColor(53, 53, 53))
    # palette.setColor(QtGui.QPalette.ToolTipText, QtCore.Qt.white)
    # palette.setColor(QtGui.QPalette.Text, QtCore.Qt.white)
    # palette.setColor(QtGui.QPalette.Button, QtGui.QColor(53, 53, 53))
    # palette.setColor(QtGui.QPalette.ButtonText, QtCore.Qt.white)
    # palette.setColor(QtGui.QPalette.BrightText, QtCore.Qt.red)
    # palette.setColor(QtGui.QPalette.Link, QtGui.QColor(42, 130, 218))
    # palette.setColor(QtGui.QPalette.Highlight, QtGui.QColor(42, 130, 218))
    # palette.setColor(QtGui.QPalette.HighlightedText, QtCore.Qt.black)
    # palette.setColor(QtGui.QPalette.Disabled,
    #                  QtGui.QPalette.Text, QtGui.QColor(164, 166, 168))
    # palette.setColor(QtGui.QPalette.Disabled,
    #                  QtGui.QPalette.WindowText, QtGui.QColor(164, 166, 168))
    # palette.setColor(QtGui.QPalette.Disabled,
    #                  QtGui.QPalette.ButtonText, QtGui.QColor(164, 166, 168))
    # palette.setColor(QtGui.QPalette.Disabled,
    #                  QtGui.QPalette.HighlightedText, QtGui.QColor(164, 166, 168))
    # palette.setColor(QtGui.QPalette.Disabled,
    #                  QtGui.QPalette.Base, QtGui.QColor(68, 68, 68))
    # palette.setColor(QtGui.QPalette.Disabled,
    #                  QtGui.QPalette.Window, QtGui.QColor(68, 68, 68))
    # palette.setColor(QtGui.QPalette.Disabled,
    #                  QtGui.QPalette.Highlight, QtGui.QColor(68, 68, 68))
    # app.setPalette(palette)

    reg_viewer = RegViewer()

    app.setWindowIcon(QtGui.QIcon(helpers.resource_path("img/icon.ico")))

    reg_viewer.show()

    # Close the splash screen
    try:
        import pyi_splash  # type: ignore
        pyi_splash.close()
    except ImportError:
        pass

    # Process command-line file(s)
    if len(sys.argv) > 1:
        for filename in sys.argv[1:]:
            reg_viewer.open_file(filename)

    sys.exit(app.exec())


if __name__ == "__main__":
    main()



import enum

import PySide6.QtCore as QtCore
import PySide6.QtWidgets as QtWidgets
from Registry import Registry

from . import helpers
from . import key_tree


class ResultType(enum.Enum):
    KEY = 0
    VALUE = 1
    DATA = 2


class FindDialog(QtWidgets.QDialog):
    def __init__(self, *args):
        super().__init__(*args)

        self.setWindowTitle("Find")
        self.resize(400, 200)
        self.setWindowFlags(QtCore.Qt.Dialog |
                            QtCore.Qt.MSWindowsFixedSizeDialogHint)

        text_group = QtWidgets.QGroupBox(self)
        text_group.setTitle("Search for")
        text_group_layout = QtWidgets.QVBoxLayout(self)
        text_group.setLayout(text_group_layout)
        self.text = QtWidgets.QLineEdit()
        self.text.setClearButtonEnabled(True)
        self.text.setToolTip(
            "For binary data, enter it in hexadecimal with spaces between bytes. E.g. 0A 54 D3")
        text_group_layout.addWidget(self.text)

        find_btn = QtWidgets.QPushButton("Find Next")
        find_btn.setDefault(True)

        self.buttonBox = QtWidgets.QDialogButtonBox()
        self.buttonBox.addButton(
            find_btn, QtWidgets.QDialogButtonBox.ButtonRole.AcceptRole)
        self.buttonBox.addButton(
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        self.buttonBox.rejected.connect(self.closed)
        self.buttonBox.accepted.connect(self.handle_find)

        options_container = QtWidgets.QWidget()
        options_container_layout = QtWidgets.QHBoxLayout()
        options_container.setLayout(options_container_layout)

        category_group = QtWidgets.QGroupBox(self)
        category_group.setTitle("Search in")
        category_group_layout = QtWidgets.QVBoxLayout(category_group)

        self.key_search = QtWidgets.QCheckBox("Keys", self)
        self.key_search.setChecked(True)
        self.value_search = QtWidgets.QCheckBox("Values", self)
        self.value_search.setChecked(True)
        self.data_search = QtWidgets.QCheckBox("Data", self)
        self.data_search.setChecked(True)

        category_group_layout.addWidget(self.key_search)
        category_group_layout.addWidget(self.value_search)
        category_group_layout.addWidget(self.data_search)
        category_group_layout.addStretch()
        options_container_layout.addWidget(category_group)

        options_group = QtWidgets.QGroupBox(self)
        options_group.setTitle("Options")
        options_group_layout = QtWidgets.QVBoxLayout(options_group)

        self.case_sensitive = QtWidgets.QCheckBox("Case Sensitive", self)
        self.exact_match = QtWidgets.QCheckBox("Exact Match", self)

        options_group_layout.addWidget(self.case_sensitive)
        options_group_layout.addWidget(self.exact_match)
        options_group_layout.addStretch()
        options_container_layout.addWidget(options_group)

        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.addWidget(text_group)
        self.layout.addWidget(options_container)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    # Override showEvent to highlight the textbox on show
    def showEvent(self, event):
        self.text.setFocus()
        super().showEvent(event)

    def handle_find(self):
        active_key: key_tree.KeyItem = self.parent().tree.get_selected_key()
        if self.text.text() == "":
            helpers.show_message_box(
                "Enter a search term first.", alert_type=helpers.MessageBoxTypes.CRITICAL)
            return

        if not (self.key_search.isChecked() or self.value_search.isChecked() or self.data_search.isChecked()):
            helpers.show_message_box(
                "You must select one type to search in.", alert_type=helpers.MessageBoxTypes.CRITICAL)
            return

        self.close()
        self.accept()

        if active_key is None:
            helpers.show_message_box(
                "Select a key or hive first.", alert_type=helpers.MessageBoxTypes.CRITICAL)
            return

        hive: Registry.Registry = self.parent().tree.reg[active_key.filename]
        current_key = hive.open(active_key.path)

        self.parent().progress_bar.show()
        self.parent().progress_bar.setRange(0, 0)
        self.parent().progress_bar.setValue(1)
        self.parent().progress_bar.setValue(0)
        result = self.find(current_key,
                           self.text.text(),
                           case_sensitive=self.case_sensitive.isChecked(),
                           exact_match=self.exact_match.isChecked(),
                           search_keys=self.key_search.isChecked(),
                           search_values=self.value_search.isChecked(),
                           search_data=self.data_search.isChecked())
        self.parent().progress_bar.setRange(0, 100)
        self.parent().progress_bar.hide()

        if result is None:
            self.parent().tree.select_key_from_path("")
            helpers.show_message_box(
                "Term not found. Looping back to start.", alert_type=helpers.MessageBoxTypes.WARNING)
            return

        result_type, result_key, result_value = result
        sanitized_path = self.parent().tree.parse_uri(
            result_key, root=hive.root().name())
        self.parent().tree.select_key_from_path(sanitized_path)
        if result_type == ResultType.VALUE or result_type == ResultType.DATA:
            self.parent().value_table.select_value(result_value)

    def find(self, starting_key: Registry.RegistryKey, term: str, case_sensitive=False, exact_match=False, search_keys=True, search_values=True, search_data=True, reverse=False) -> "tuple[ResultType, str, str]":
        """Find the next matching subkey or value. Returns (ResultType, key, value)"""

        if not case_sensitive:
            term = term.upper()

        def check_match(text: str) -> bool:
            """Check if a subkey matches the search term"""
            if not case_sensitive:
                text = text.upper()
            if exact_match:
                if term == text:
                    return True
            else:
                if term in text:
                    return True

            return False

        def search(start_key: Registry.RegistryKey, term: str, start_at_value=0, skip_start_key_name=False) -> "tuple[ResultType, str, str]":
            """Returns (ResultType, key, value)"""

            # Check the start key name if asked (i.e. if the search has just started)
            if search_keys and not skip_start_key_name:
                if check_match(start_key.name()):
                    return ResultType.KEY, start_key.path(), None

            values = start_key.values()[start_at_value:]
            # Skip extra looping if values and data are not searched for
            if search_values or search_data:
                for value in values:
                    # Check through the value
                    if search_values and check_match(value.name()):
                        return ResultType.VALUE, start_key.path(), value.name()
                    # Check through the value's data
                    if search_data and (
                            check_match(str(value.value())) or
                            check_match(self.parent().value_table.reg_data_to_str(value.value_type(), value.raw_data(), value.value()))):
                        return ResultType.DATA, start_key.path(), value.name()

            # Recurse through the subkeys
            for subkey in start_key.subkeys():
                # Check subkey name
                if search_keys and check_match(subkey.name()):
                    return ResultType.KEY, subkey.path(), None
                result = search(subkey, term)
                if result is not None:
                    # Found a recursive match!
                    return result

            # No match was found
            return None

        match = search(starting_key, term,
                       start_at_value=self.parent().value_table.get_selected_row() + 1, skip_start_key_name=True)
        if match is not None:
            return match

        current_key = starting_key
        try:
            while current_key.parent():
                subkeys = current_key.parent().subkeys()
                index = next((i for i, subkey in enumerate(subkeys)
                             if subkey.name() == current_key.name()))
                subkeys = subkeys[index+1:]
                for subkey in subkeys:
                    result = search(subkey, term)
                    if result is not None:
                        return result
                current_key = current_key.parent()
        except Registry.RegistryKeyHasNoParentException:
            pass

        return None

    def closed(self):
        self.close()
        self.reject()



import os
import sys

import PySide6.QtWidgets as QtWidgets


APP_NAME = "Registry Spy"
VERSION = (1, 0, 2)
ORGANIZATION_NAME = "Andy Smith"
ORGANIZATION_DOMAIN = "ajsmith.us"
ABOUT_TEXT = f"""\
{APP_NAME}
{".".join(str(i) for i in VERSION)}
Copyright (C) 2021 Andy Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


def resource_path(relative_path: str) -> str:
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.dirname(os.path.realpath(__file__))

    return os.path.join(base_path, relative_path)


class MessageBoxTypes:
    INFORMATION = (QtWidgets.QMessageBox.Information, "Information")
    WARNING = (QtWidgets.QMessageBox.Warning, "Warning")
    CRITICAL = (QtWidgets.QMessageBox.Critical, "Error")
    QUESTION = (QtWidgets.QMessageBox.Question, "Question")


def show_message_box(text, alert_type=MessageBoxTypes.INFORMATION, title=None):
    msgbox = QtWidgets.QMessageBox()
    if title is None:
        msgbox.setWindowTitle(alert_type[1])
    msgbox.setText(text)
    msgbox.setIcon(alert_type[0])
    return msgbox.exec()



import struct

from Registry import Registry
import PySide6.QtCore as QtCore
import PySide6.QtWidgets as QtWidgets
import PySide6.QtGui as QtGui

from . import helpers


class KeyItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, path: str, filename: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = path
        self.filename = filename


class KeyTree(QtWidgets.QTreeWidget):
    """Tree widget that displays registry keys"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.roots: dict[str, KeyItem] = {}
        self.reg: dict[str, Registry.Registry] = {}

        self.setColumnCount(3)
        self.setHeaderLabels(["Key", "Subkeys", "Modified"])
        self.header().setStretchLastSection(False)
        self.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        self.header().setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
        self.header().setSectionResizeMode(2, QtWidgets.QHeaderView.ResizeToContents)
        self.selectionModel().selectionChanged.connect(self.handle_selection_change)
        self.expanded.connect(self.handle_expand)
        self.collapsed.connect(self.handle_collapse)

        self.key_icon = QtGui.QIcon(
            helpers.resource_path("img/folder.png"))
        self.hive_icon = QtGui.QIcon(
            helpers.resource_path("img/icon.png"))

    def get_uri_textbox(self) -> QtWidgets.QLineEdit:
        """Returns the URI textbox from the main form"""
        return self.window().uri_textbox

    def get_selected_key(self) -> KeyItem:
        """Returns the selected key in the KeyTree"""
        selected = self.selectedItems()
        if len(selected) > 0:
            return selected[0]

    def get_selected_hive(self) -> KeyItem:
        """Returns the root KeyItem of the selected KeyItem"""
        key = self.get_selected_key()
        if key is None:
            helpers.show_message_box(
                "No key selected, select a key first.", alert_type=helpers.MessageBoxTypes.CRITICAL)
            return
        if self.roots.get(key.filename) is not None:
            return self.roots[key.filename]

    def remove_all_hives(self):
        """Unload all hives"""
        for root in list(self.roots.values()):
            self.remove_hive(root)

    def remove_selected_hive(self):
        """Unload the selected hive"""
        root = self.get_selected_hive()
        if root is None:
            return
        self.remove_hive(root)

    def remove_hive(self, root: KeyItem):
        """Unloads a hive specified by the provided root KeyItem"""
        self.window().hive_info.set_info("", "", "", "")

        filename = root.filename
        self.takeTopLevelItem(self.indexOfTopLevelItem(root))
        self.window().value_table.set_data([])
        self.get_uri_textbox().setText("")

        del self.roots[filename]
        del self.reg[filename]

    def load_hive(self, filename: str):
        """Load a registry hive from a file"""

        # Remove hive before loading another one
        if self.roots.get(filename) is not None:
            helpers.show_message_box(
                "Registry hive already open, close it first before opening again", alert_type=helpers.MessageBoxTypes.CRITICAL)
            return

        try:
            self.reg[filename] = Registry.Registry(filename)
        except (Registry.RegistryParse.ParseException, struct.error):
            helpers.show_message_box(
                "Unable to parse registry file", alert_type=helpers.MessageBoxTypes.CRITICAL)
            return

        # Create new root KeyItem
        self.roots[filename] = KeyItem(
            "", filename, [f"{self.reg[filename].hive_type().name} ({filename})", str(self.reg[filename].root().subkeys_number()), self.reg[filename].root().timestamp().strftime("%Y-%m-%d %H:%M:%S")])

        self.roots[filename].setIcon(0, self.hive_icon)
        self.load_subkeys(self.roots[filename])

        self.insertTopLevelItems(0, [self.roots[filename]])
        self.clearSelection()
        self.roots[filename].setSelected(True)
        self.setFocus()

    def set_uri(self, index: QtCore.QModelIndex):
        """Set navbar full key path"""
        if isinstance(self.itemFromIndex(index), KeyItem):
            path = self.format_uri(self.itemFromIndex(index))
            self.get_uri_textbox().setText(path)

    def format_uri(self, key: KeyItem) -> str:
        """Format a URI path for the specified KeyItem"""
        if key.path == "":
            return self.reg[key.filename].hive_type().name
        else:
            return self.reg[key.filename].hive_type().name + "\\" + key.path

    def parse_uri(self, uri: str, hive_type: str = None, root: str = None) -> str:
        """Parses a user-specified URI into a registry path"""

        # Sanitize URI
        uri = uri.strip()
        uri = uri.strip("\\")
        if hive_type is not None:
            uri = uri.replace(hive_type + "\\", "", 1)
            uri = uri.strip("\\")
        if root is not None:
            uri = uri.replace(root + "\\", "", 1)
            uri = uri.strip("\\")

        if uri == "" or uri == hive_type:
            return ""

        return uri

    def remove_hive_prefix(self, root_name: str, path: str) -> str:
        prefix = root_name + "\\"
        return path.replace(prefix, "", 1)

    def load_subkeys(self, key: KeyItem):
        for i in range(key.childCount()):
            key.removeChild(key.child(0))

        root_name = self.reg[key.filename].root().name()

        if key.childCount() > 1:
            return

        num_subkeys = self.reg[key.filename].open(key.path).subkeys_number()

        # Don't bother showing the progress bar if there aren't many items
        display_progressbar = num_subkeys > 20

        if display_progressbar:
            self.window().progress_bar.setMaximum(num_subkeys)
            self.window().progress_bar.setValue(0)
            self.window().progress_bar.show()
            i = 0

        for subkey in self.reg[key.filename].open(key.path).subkeys():
            if display_progressbar:
                # Process events once in a while so the application doesn't "stop responding" on Windows
                if i % 20 == 0:
                    self.window().progress_bar.setValue(i)
                # if i % 1000 == 0:
                #     QtCore.QCoreApplication.processEvents()
                # Removed for now because processEvents causes a TON of delay
                i += 1

            subkey_child = KeyItem(self.remove_hive_prefix(root_name, subkey.path()), key.filename, [
                                   subkey.name(), str(subkey.subkeys_number()), subkey.timestamp().strftime("%Y-%m-%d %H:%M:%S")])
            subkey_child.setIcon(0, self.key_icon)
            key.addChild(subkey_child)

            # Create a fake child so that the tree shows an arrow to drop down
            if subkey.subkeys_number() > 0:
                subkey_child.addChild(KeyItem("", ""))

            # for subsubkey in subkey.subkeys():
            #     subsubkey_child = KeyItem(
            #         self.remove_hive_prefix(root_name, subsubkey.path()), subkey_child.filename, [subsubkey.name()])
            #     subkey_child.addChild(subsubkey_child)

        if display_progressbar:
            self.window().progress_bar.hide()

    def select_key_from_path(self, path: str) -> KeyItem:
        """Find a KeyItem from a given path and highlight it"""
        parent = self.get_selected_hive()
        if parent is None:
            return
        levels = path.count("\\") + 1

        # Check if root is selected
        if path == parent.path:
            self.clearSelection()
            self.scrollToItem(parent)
            parent.setSelected(True)
            self.setFocus()
            return

        for level in range(levels):
            # Expand parent
            parent.setExpanded(True)
            for c in range(parent.childCount()):
                subpath = "\\".join(path.split("\\")[:level+1])
                child_path = parent.child(c).path
                # Last level, should have match
                if child_path == subpath:
                    if level + 1 == levels:
                        # Found!
                        self.clearSelection()
                        self.scrollToItem(parent.child(c))
                        parent.child(c).setSelected(True)
                        self.setFocus()
                        return parent.child(c)

                    # Set parent to new child
                    parent = parent.child(c)
                    break

    def handle_uri_change(self):
        root = self.get_selected_hive()
        if root is None:
            return

        # Get the new text
        uri: str = self.get_uri_textbox().text()
        parsed_uri = self.parse_uri(
            uri, self.reg[root.filename].hive_type().name)

        try:
            self.reg[root.filename].open(parsed_uri)
        except Registry.RegistryKeyNotFoundException:
            helpers.show_message_box(
                "Key was not found", alert_type=helpers.MessageBoxTypes.CRITICAL)
            return

        key = self.select_key_from_path(parsed_uri)

    def handle_selection_change(self, selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection):
        if len(selected.indexes()) < 1:
            return

        index = selected.indexes()[0]

        self.set_uri(index)

        key: KeyItem = self.itemFromIndex(index)

        self.window().hive_info.set_info(key.filename, self.reg[key.filename].hive_type(
        ).name, self.reg[key.filename].hive_name(), self.reg[key.filename].root().name())

        try:
            self.window().value_table.set_data(
                self.reg[key.filename].open(key.path).values())
        except Registry.RegistryKeyNotFoundException:
            self.window().value_table.set_data([])

    def handle_expand(self, index: QtCore.QModelIndex):
        self.window().statusBar().showMessage("Loading...")
        self.window().statusBar().repaint()
        self.setCursor(QtCore.Qt.CursorShape.BusyCursor)

        key = self.itemFromIndex(index)
        self.load_subkeys(key)

        self.window().statusBar().clearMessage()
        self.unsetCursor()

    def handle_collapse(self, index: QtCore.QModelIndex):
        key = self.itemFromIndex(index)
        for i in range(key.childCount()):
            key.removeChild(key.child(0))
        key.addChild(KeyItem("", ""))



import PySide6.QtWidgets as QtWidgets
import PySide6.QtCore as QtCore


class HiveInfoTable(QtWidgets.QTableWidget):
    """Value table that shows the values of the selected registry key"""

    def __init__(self, parent, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setParent(parent)
        self.setColumnCount(1)

        header_labels = ["Path", "Type", "Hive Name", "Root Name"]
        self.setRowCount(len(header_labels))
        self.setVerticalHeaderLabels(header_labels)

        self.setSizePolicy(QtWidgets.QSizePolicy.Ignored,
                           QtWidgets.QSizePolicy.Maximum)
        self.setMinimumHeight(0)
        self.horizontalHeader().setStretchLastSection(True)
        self.setSelectionMode(
            QtWidgets.QAbstractItemView.SelectionMode.SingleSelection)
        self.horizontalHeader().setVisible(False)
        self.resizeRowsToContents()
        self.setAutoScroll(False)
        self.setVerticalScrollMode(
            QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setHorizontalScrollMode(
            QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setWordWrap(False)
        self.verticalHeader().setSectionResizeMode(self.verticalHeader().Fixed)

        # Call so that the cells become uneditable
        self.set_info("", "", "", "")

    def set_info(self, hive_path, hive_type, hive_name, root_name):
        """Set the hive info table data"""

        hive_path_item = QtWidgets.QTableWidgetItem(hive_path)
        hive_path_item.setFlags(hive_path_item.flags() & ~
                                QtCore.Qt.ItemFlag.ItemIsEditable)
        self.setItem(0, 0, hive_path_item)

        hive_type_item = QtWidgets.QTableWidgetItem(hive_type)
        hive_type_item.setFlags(hive_type_item.flags() & ~
                                QtCore.Qt.ItemFlag.ItemIsEditable)
        self.setItem(0, 1, hive_type_item)

        hive_name_item = QtWidgets.QTableWidgetItem(hive_name)
        hive_name_item.setFlags(hive_name_item.flags() & ~
                                QtCore.Qt.ItemFlag.ItemIsEditable)
        self.setItem(0, 2, hive_name_item)

        root_name_item = QtWidgets.QTableWidgetItem(root_name)
        root_name_item.setFlags(root_name_item.flags() & ~
                                QtCore.Qt.ItemFlag.ItemIsEditable)
        self.setItem(0, 3, QtWidgets.QTableWidgetItem(root_name_item))



import PySide6.QtCore as QtCore
import PySide6.QtWidgets as QtWidgets

from . import helpers


class LicenseDialog(QtWidgets.QDialog):
    def __init__(self, *args):
        super().__init__(*args)

        self.setWindowTitle("Third-Party Licenses")
        self.resize(550, 400)

        text = QtWidgets.QPlainTextEdit()
        text.setReadOnly(True)
        text_file = QtCore.QFile(
            helpers.resource_path("third_party_licenses.txt"))
        if not text_file.open(QtCore.QIODevice.ReadOnly):
            print(text_file.errorString())

        stream = QtCore.QTextStream(text_file)
        text.setPlainText(stream.readAll())

        ok_btn = QtWidgets.QDialogButtonBox.StandardButton.Ok

        self.buttonBox = QtWidgets.QDialogButtonBox(ok_btn)
        self.buttonBox.accepted.connect(self.accept)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(text)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="registryspy",
    version="1.0.2",
    author="Andy Smith",
    description="Cross-platform Windows Registry browser",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/andyjsmith/Registry-Spy",
    keywords=[
        "registry",
        "forensics",
        "windows forensics",
        "forensics tools"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Environment :: Win32 (MS Windows)",
        "Environment :: X11 Applications",
        "Environment :: X11 Applications :: Qt",
        "Environment :: MacOS X"
    ],
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "PySide6",
        "python-registry"
    ],
    entry_points={
        "console_scripts": ["registryspy=registryspy.registryspy:main"],
    },
    include_package_data=True
)



import struct

from Registry import Registry
import PySide6.QtWidgets as QtWidgets
import PySide6.QtGui as QtGui
import PySide6.QtCore as QtCore

from . import helpers


class ValueData(QtWidgets.QTableWidgetItem):
    def __init__(self, raw_data: bytes, value, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.raw_data = raw_data
        self.value = value


class ValueTable(QtWidgets.QTableWidget):
    """Value table that shows the values of the selected registry key"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setColumnCount(3)
        self.setHorizontalHeaderLabels(["Name", "Type", "Data"])
        self.horizontalHeader().setStretchLastSection(True)
        self.setColumnWidth(0, 180)
        self.setColumnWidth(1, 120)
        self.setSelectionMode(
            QtWidgets.QAbstractItemView.SelectionMode.SingleSelection)
        self.setSelectionBehavior(
            QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.data = None
        self.verticalHeader().setVisible(False)
        self.horizontalHeader().setDefaultAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
        self.resizeRowsToContents()
        self.setAutoScroll(False)
        self.setVerticalScrollMode(
            QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setHorizontalScrollMode(
            QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setWordWrap(False)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)

        self.selectionModel().selectionChanged.connect(self.handle_selection_change)

        self.REG_BIN_ICON = QtGui.QIcon(
            helpers.resource_path("img/reg_bin.png"))
        self.REG_NUM_ICON = QtGui.QIcon(
            helpers.resource_path("img/reg_num.png"))
        self.REG_STR_ICON = QtGui.QIcon(
            helpers.resource_path("img/reg_str.png"))

    def handle_selection_change(self, selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection):
        if len(selected.indexes()) < 1:
            return

        index = selected.indexes()[2]

        value: ValueData = self.itemFromIndex(index)

        hex_text = ""

        def chunks(lst, n):
            """Yield successive n-sized chunks from lst."""
            for i in range(0, len(lst), n):
                yield lst[i:i + n]

        for chunk in chunks(value.raw_data, 16):
            hex_text += " ".join(["{:02x}".format(x)
                                  for x in chunk]) + "\n"

        self.window().value_hex.setPlainText(hex_text)

    def set_data(self, reg_values: "list[Registry.RegistryValue]"):
        self.clearContents()
        self.setRowCount(0)
        self.window().value_hex.setPlainText("")

        for value in reg_values:
            index = self.rowCount()
            self.insertRow(index)
            value_name = QtWidgets.QTableWidgetItem(value.name())
            value_name.setFlags(value_name.flags() & ~
                                QtCore.Qt.ItemFlag.ItemIsEditable)
            value_name.setIcon(self.get_icon(value.value_type()))
            value_type = QtWidgets.QTableWidgetItem(
                self.reg_type_to_str(value.value_type()))
            value_type.setFlags(value_type.flags() & ~
                                QtCore.Qt.ItemFlag.ItemIsEditable)
            value_data = ValueData(value.raw_data(), value.value(), self.reg_data_to_str(
                value.value_type(), value.raw_data(), value.value()))
            if len(value.raw_data()) == 0:
                font = QtGui.QFont()
                font.setItalic(True)
                value_data.setFont(font)
            value_data.setFlags(value_data.flags() & ~
                                QtCore.Qt.ItemFlag.ItemIsEditable)
            self.setItem(index, 0, value_name)
            self.setItem(index, 1, value_type)
            self.setItem(index, 2, value_data)

        self.resizeRowsToContents()

    def get_icon(self, datatype: int) -> QtGui.QIcon:
        if datatype == Registry.RegBin:
            return self.REG_BIN_ICON
        if datatype == Registry.RegDWord:
            return self.REG_NUM_ICON
        if datatype == Registry.RegQWord:
            return self.REG_NUM_ICON
        if datatype == Registry.RegBigEndian:
            return self.REG_NUM_ICON
        if datatype == Registry.RegExpandSZ:
            return self.REG_STR_ICON
        if datatype == Registry.RegLink:
            return self.REG_STR_ICON
        if datatype == Registry.RegMultiSZ:
            return self.REG_STR_ICON
        if datatype == Registry.RegNone:
            return self.REG_NUM_ICON
        if datatype == Registry.RegResourceList:
            return self.REG_STR_ICON
        if datatype == Registry.RegSZ:
            return self.REG_STR_ICON
        return self.REG_BIN_ICON

    def reg_type_to_str(self, datatype: int) -> str:
        if datatype == Registry.RegBin:
            return "REG_BINARY"
        if datatype == Registry.RegDWord:
            return "REG_DWORD"
        if datatype == Registry.RegQWord:
            return "REG_QWORD"
        if datatype == Registry.RegBigEndian:
            return "REG_DWORD_BIG_ENDIAN"
        if datatype == Registry.RegExpandSZ:
            return "REG_EXPAND_SZ"
        if datatype == Registry.RegLink:
            return "REG_LINK"
        if datatype == Registry.RegMultiSZ:
            return "REG_MULTI_SZ"
        if datatype == Registry.RegNone:
            return "REG_NONE"
        if datatype == Registry.RegResourceList:
            return "REG_RESOURCE_LIST"
        if datatype == Registry.RegSZ:
            return "REG_SZ"
        return "UNKNOWN"

    def reg_data_to_str(self, datatype: int, raw_data: bytes, value) -> str:
        if len(raw_data) == 0:
            return "(value not set)"
        if datatype == Registry.RegDWord:
            try:
                return "{0:#010x} ({0})".format(value)
            except (struct.error, IndexError):
                pass
        if datatype == Registry.RegQWord:
            try:
                return "{0:#018x} ({0})".format(value)
            except (struct.error, IndexError):
                pass
        if datatype == Registry.RegBigEndian:
            try:
                return "{0:#010x} ({0})".format(value)
            except (struct.error, IndexError):
                pass
        if datatype == Registry.RegLink:
            # Not sure what format this will actually be
            return str(value)
        if datatype == Registry.RegMultiSZ:
            return " ".join(value)
        if datatype == Registry.RegResourceList:
            # Not sure what format this will actually be
            return str(value)
        if datatype == Registry.RegSZ or datatype == Registry.RegExpandSZ:
            return value
        else:
            return " ".join(["{:02x}".format(x) for x in raw_data])

    def select_value(self, value: str):
        for i in range(self.rowCount()):
            if (value == self.item(i, 0).text()):
                self.clearSelection()
                self.selectRow(i)
                self.scrollToItem(self.item(i, 0))
                self.setFocus()

    def get_selected_row(self):
        if len(self.selectedItems()) > 0:
            return self.selectedItems()[0].row()
        else:
            return -1
